--Handles all the syncing
--Basic concept is
--1. Only send messages for yoursel
--2. Send the event number, if someone receives an event number higher than their current number then they missed the event
--3. Reset event numbers shortly after combat ends (delayed in case messages arrive late)
--4. Use number abbreviations for text (abbreviations are first registered, anyone can send a message requesting all abbreviations be resent as they get used in case of disconnects)
--5. Queue messages together to send at once 


local SyncMsg=false

local MaxMsgs=12
local MsgNum=0
local MsgQueue={}

local AbbrevNum=0	--How many phrases in the abbreviation table
local AbbrevTable={}	--Converts phrases to numbers
local RAbbrevTable={}	--Converts numbers to phrases

local RangeSync={}

--This tells us what args in the tables should need conversion
local NeedsConvert={
	DAMAGE={true,true,true,true,true,false,false},
	HEALING={true,true,true,true,false,false}
}

local function GetAbbrev(phrase)
	if not AbbrevTable[phrase] then
		local Abbrev
		AbbrevNum=AbbrevNum+1

		Abbrev=AbbrevNum

		--Lets make full use of the serialization spec by negating if needed
		if Abbrev>127 and Abbrev<=256 then
			Abbrev=Abbrev-256
		elseif Abbrev>256 then
			Abbrev=Abbrev-129
			--Shouldn't ever get to 32767 abbreviations so won't bother with that
		end
		AbbrevTable[phrase]=AbbrevNum

		if not RAbbrevTable[Recount.PlayerName] then
			RAbbrevTable[Recount.PlayerName]=Recount:GetTable()
		end

		RAbbrevTable[Recount.PlayerName][AbbrevNum]=phrase
		
		--Need to add a message now
		MsgNum=MsgNum+1

		MsgQueue[MsgNum]=Recount:GetTable()
		MsgQueue[MsgNum][1]="ABBREV"
		MsgQueue[MsgNum][2]=AbbrevNum
		MsgQueue[MsgNum][3]=phrase

		if MsgNum==MaxMsgs then
			Recount:SendMsgQueue()
		end

		return AbbrevNum
	end
	return AbbrevTable[phrase]
end

--Replaces Abbreviations with the proper phrases
--If unable to translate due to missing phrases it will return false otherwise true
local function TranslateArgTable(Who, EventType, ArgTable)
	local TransTable=RAbbrevTable[Who]
	if type(TransTable)~="table" then
		return false
	end

	local ArgTypes=NeedsConvert[EventType]

	for k, v in pairs(ArgTable) do
		if ArgTypes[k] then
			if TransTable[v] then
				ArgTable[k]=TransTable[v]
			else
				return false
			end
		end
	end

	return true
end

function Recount:SendEvent(source, victim, EventType, argTable)	
	--Make sure this isn't a request to add an event from a sync message
	if SyncMsg then
		Recount:FreeTable(argTable)
		return
	end

	--Inc the event numbers
	Recount.EventNum[EventType][source]=(Recount.EventNum[EventType][source] or 0)+1	

	--Only send messages for ourself
	if source~=Recount.PlayerName and victim~=Recount.PlayerName then
		--Except if we are the closest
		if  (not RangeSync.CanSendFor(source)) and (not RangeSync.CanSendFor(victim)) then
			Recount:FreeTable(argTable)
			return
		end
	end

	--First thing the argTable needs to be converted
	for k, v in pairs(argTable) do
		if NeedsConvert[EventType][k] then
			argTable[k]=GetAbbrev(v)
		end
	end

	MsgNum=MsgNum+1

	MsgQueue[MsgNum]=Recount:GetTable()
	MsgQueue[MsgNum][1]=EventType
	MsgQueue[MsgNum][2]=Recount.EventNum[EventType][source]
	MsgQueue[MsgNum][3]=argTable

	if MsgNum==MaxMsgs then
		Recount:SendMsgQueue()
	end
end

function Recount:ResetAbbrev()
	for k in pairs(AbbrevTable) do
		AbbrevTable[k]=nil
	end
	AbbrevNum=0
end

function Recount:RequestResetAbbrev()
	Recount:ResetAbbrev()

	MsgNum=MsgNum+1

	MsgQueue[MsgNum]=Recount:GetTable()
	MsgQueue[MsgNum][1]="RESETABBREV"
	Recount:SendMsgQueue()
end


function Recount:RequestVersion()
	MsgNum=MsgNum+1

	MsgQueue[MsgNum]=Recount:GetTable()
	MsgQueue[MsgNum][1]="VERSIONREQUEST"
end


function Recount:SendVersion()
	MsgNum=MsgNum+1

	MsgQueue[MsgNum]=Recount:GetTable()
	MsgQueue[MsgNum][1]="VERSION"
	MsgQueue[MsgNum][2]=Recount.Version
end

function Recount:SendMsgQueue()
	if MsgNum==0 then
		return
	end

	local result=Recount:SendCommMessage("GROUP",MsgQueue)

	MsgNum=0
	for k,v in pairs(MsgQueue) do
		Recount:FreeTableRecurse(v)
		MsgQueue[k]=nil
	end
end

function Recount:DumpMsgQueue()
	Recount:PrintLiteral(MsgQueue)
end

function Recount:DumpAbbrevTable(other)
	if other==nil then
		other=Recount.PlayerName
	end
	local OurTable=RAbbrevTable[other]
	Recount:Print(AbbrevNum.." Abbreviations")

	for i=1,AbbrevNum do 
		Recount:Print("Abbrev for "..OurTable[i].." is "..i)
	end
end

function Recount:OnCommReceive(prefix, sender, distribution, Msgs)
	local EventType
	
	--Make sure we received a table and right prefix
	if type(Msgs)~="table" then
		return
	end

	SyncMsg=true
	for k, v in pairs(Msgs) do
		EventType=v[1]
		if EventType=="ABBREV" then	--Add Abbreviation
			if type(RAbbrevTable[sender])~="table" then
				RAbbrevTable[sender]=Recount:GetTable()
			end

			RAbbrevTable[sender][v[2]]=v[3]
		elseif EventType=="RESETABBREV" then	--Reset abbreviations being used (done for joining groups)
			for k in pairs(AbbrevTable) do
				AbbrevTable[k]=nil
			end
			AbbrevNum=0
		elseif EventType=="DAMAGE" then
			if TranslateArgTable(sender, EventType, v[3]) then
				--Is the event number less
				if (Recount.EventNum[EventType][v[3][1]] or 0)<v[2] then					
					Recount.EventNum[EventType][v[3][1]]=v[2]
					Recount:AddDamageData(v[3][1],v[3][2],v[3][3],v[3][4],v[3][5],v[3][6],v[3][7])
				end
			end
		elseif EventType=="HEALING" then
			if TranslateArgTable(sender, EventType, v[3]) then
				--Is the event number less
				if (Recount.EventNum[EventType][v[3][1]] or 0)<v[2] then
					Recount.EventNum[EventType][v[3][1]]=v[2]
					Recount:AddHealData(v[3][1],v[3][2],v[3][3],v[3][4],v[3][5],v[3][6])
				end
			end
		elseif EventType=="VERSIONREQUEST" then
			Recount:SendVersion()
		elseif EventType=="VERSION" then
			Recount.VerTable[sender]=v[2]
		elseif EventType=="RANGELIST" then
			RangeSync.ReceiveRangeList(sender,v[2])
		end
	end
	SyncMsg=false
end



--Range checking for sending other player events
local rc = AceLibrary("RangeCheck-1.0")
local Roster = AceLibrary("Roster-2.1")

--Current self range list being used
local PlayerRangeList={}

local CheckNext
local PlayerCheckList={}
local PlayerUnitList={}
local NewPlayerRangeList={}

local RangeLists={}
local LastUpdated={}



function RangeSync.DeterminePlayerCheckList()
	for k in pairs(PlayerCheckList) do
		PlayerCheckList[k]=nil
		PlayerUnitList[k]=nil
	end

	for k in pairs(NewPlayerRangeList) do
		NewPlayerRangeList[k]=nil
	end	

	for unit in Roster:IterateRoster(true) do
		PlayerCheckList[#PlayerCheckList+1]=unit.name
		PlayerUnitList[#PlayerUnitList+1]=unit.unitid
	end
	
	if #PlayerCheckList==0 then
		PlayerCheckList[1]=UnitName("player")
		PlayerUnitList[1]="player"
	end
	
	--Start at 1
	CheckNext=1

	--Want to send a new list ~10 seconds so update the scheduled event repeat time for this
	Recount:ScheduleRepeatingEvent("SyncRangeUpdate",RangeSync.SyncRangeUpdate,15/#PlayerCheckList)
end

function RangeSync.UpdateNextRanged()
	local name = PlayerCheckList[CheckNext]
	local unit = PlayerUnitList[CheckNext]

	if not UnitExists(unit) or UnitName(unit)~=name then
		unit = Roster:GetUnitIDFromName(name)

		--Make sure a unit was actually found
		if not unit then
			CheckNext=CheckNext+1
			return
		end
	end

	local _, max = rc:getRange(unit)

	if not max then
		max=100
	end

	if (not PlayerRangeList[name]) or (PlayerRangeList[name]>max) then
		PlayerRangeList[name]=max
	end

	if (not NewPlayerRangeList[name]) or (NewPlayerRangeList[name]>max) then
		NewPlayerRangeList[name]=max
	end

	CheckNext=CheckNext+1
end

function RangeSync.SyncRangeUpdate()
	RangeSync.UpdateNextRanged()

	--Have we checked all the players?
	if CheckNext>#PlayerCheckList then
		--Swap the new list with the old one
		local T=PlayerRangeList
		PlayerRangeList=NewPlayerRangeList
		NewPlayerRangeList=T

		PlayerRangeList[Recount.PlayerName]=0

		--Send the list
		RangeSync.SendRangeList()

		--Remove any lists still around that shouldn't be there
		RangeSync.RemoveOldLists()

		--Create the new potential player list
		RangeSync.DeterminePlayerCheckList()
	end
end

function RangeSync.SendRangeList()
	local SendList=Recount:GetTable()
	local Abbrev,i

	i=1

	for k, v in pairs(PlayerRangeList) do 
		SendList[i]=GetAbbrev(k)
		SendList[i+1]=v
		i=i+2
	end

	MsgNum=MsgNum+1

	MsgQueue[MsgNum]=Recount:GetTable()
	MsgQueue[MsgNum][1]="RANGELIST"
	MsgQueue[MsgNum][2]=SendList
	
	--Want this to go now
	Recount:SendMsgQueue()
end

function Recount:TestRangeReceive()
	local SendList=Recount:GetTable()
	local Abbrev,i

	i=1

	for k, v in pairs(PlayerRangeList) do 
		SendList[i]=GetAbbrev(k)
		SendList[i+1]=v
		i=i+2
	end
	
	RangeSync.ReceiveRangeList(Recount.PlayerName,SendList)
end

function RangeSync.ReceiveRangeList(who,list)
	LastUpdated[who]=GetTime()

	local TransTable=RAbbrevTable[who]
	--make sure everything is in the translation table (in case of a reset before the msg was received)
	for i=1,#list,2 do
		if not TransTable or not TransTable[list[i]] then
			return
		end		
	end

	if not RangeLists[who] then
		RangeLists[who]=Recount:GetTable()
	end

	local RList=RangeLists[who]

	for k in pairs(RList) do
		RList[k]=nil
	end

	local Who
	

	for i=1,#list,2 do
		Who=TransTable[list[i]]
		RList[Who]=list[i+1]
	end
end

function RangeSync.RemoveOldLists()
	local OutdatedTime=GetTime()-25
	for k, v in pairs(LastUpdated) do
		if v<OutdatedTime then
			Recount:FreeTable(RangeLists[k])
			RangeLists[k]=nil
			LastUpdated[k]=nil
		end
	end
end

local ThrottleAt=0

function RangeSync.CanSendFor(who)
	local OurRange=PlayerRangeList[who]

	if OurRange==nil then
		return
	end

	
	

	for k, v in pairs(RangeLists) do
		if v[who] and (v[who]<=OurRange or (v[who]==OurRange)) then
			return false
		end
	end

	if ThrottleAt>=32 then
		return false
	end
	ThrottleAt=ThrottleAt+1

	return true
end

function Recount:DecThrottle()
	if ThrottleAt>0 then
		ThrottleAt=ThrottleAt-1
	end
end



function Recount:DumpRangeList()
	Recount:Print("Range List")
	for k, v in pairs(PlayerRangeList) do
		Recount:Print(k.." "..v)
	end
end

function Recount:DumpOtherRangeList()
	for k, v in pairs(RangeLists) do
		Recount:Print("Range List for "..k)
		for k2, v2 in pairs(v) do
			Recount:Print(k2.." "..v2)
		end
	end
end


--Need this at the end
function Recount:ConfigComm()
	Recount:SetCommPrefix("RECOUNT")
	Recount:RegisterComm("RECOUNT", "GROUP", "OnCommReceive")
	Recount:SetDefaultCommPriority("ALERT")
	Recount:RegisterMemoizations("ABBREV","RESETABBREV","DAMAGE","HEALING","VERSIONREQUEST","VERSION","RANGELIST")

	Recount:ScheduleRepeatingEvent("RecountTick",Recount.DecThrottle,0.1)

	Recount.VerTable={}

	--Need to request a reset in case we DC'ed and rejoined into a raid
	Recount:SendVersion()
	Recount:RequestVersion()
	Recount:RequestResetAbbrev()

	--Setup Range Checking
	RangeSync.DeterminePlayerCheckList()
end
